<?php

namespace Drupal\eav_field\Form;

use Drupal\Core\Entity\EntityFieldManagerInterface;
use Drupal\Core\Entity\EntityInterface;
use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\Core\Field\FieldDefinitionInterface;
use Drupal\Core\Field\WidgetInterface;
use Drupal\Core\Field\WidgetPluginManager;
use Drupal\Core\Form\FormBase;
use Drupal\Core\Form\FormStateInterface;
use Symfony\Component\DependencyInjection\ContainerInterface;

class EavEntityAttributesForm extends FormBase {

  /**
   * @var EntityTypeManagerInterface
   */
  protected $entityTypeManager;

  /**
   * @var EntityFieldManagerInterface
   */
  protected $entityFieldManager;

  /**
   * @var WidgetPluginManager
   */
  protected $fieldWidgetManager;

  /**
   * Form constructor.
   */
  public function __construct(EntityTypeManagerInterface $entity_type_manager, EntityFieldManagerInterface $entity_field_manager, WidgetPluginManager $field_widget_manager) {
    $this->entityTypeManager = $entity_type_manager;
    $this->entityFieldManager = $entity_field_manager;
    $this->fieldWidgetManager = $field_widget_manager;
  }

  /**
   * {@inheritDoc}
   */
  public static function create(ContainerInterface $container) {
    return new static(
      $container->get('entity_type.manager'),
      $container->get('entity_field.manager'),
      $container->get('plugin.manager.field.widget')
    );
  }

  /**
   * {@inheritDoc}
   */
  public function getFormId() {
    return 'eav_entity_attributes_form';
  }

  /**
   * {@inheritDoc}
   */
  public function buildForm(array $form, FormStateInterface $form_state) {
    $entity = $this->getEntityFromRoute();
    $eav_field_definition = $this->getEntityEavFieldDefinition($entity);
    $eav_field_name = $eav_field_definition->getName();

    $eav_widget = $this->fieldWidgetManager->getInstance([
      'field_definition' => $eav_field_definition,
      'configuration' => [
        'type' => 'eav_widget',
      ],
    ]);

    $form['#parents'] = $form['#parents'] ?? [];
    $form[$eav_field_name] = $eav_widget->form($entity->get($eav_field_name), $form, $form_state);

    $form['actions'] = [
      '#type' => 'actions',
    ];

    $form['actions']['save'] = [
      '#type' => 'submit',
      '#value' => $this->t('Save'),
      '#button_type' => 'primary',
    ];

    if (!$entity->get($eav_field_name)->isEmpty()) {
      $form['actions']['delete'] = [
        '#type' => 'submit',
        '#value' => $this->t('Clear all attributes value'),
        '#button_type' => 'danger',
        '#submit' => ['::clearAllValues'],
      ];
    }

    $form['#title'] = $this->t('Editing attributes of "@entity_label"', ['@entity_label' => $entity->label()]);

    $form_state->set('entity', $entity);
    $form_state->set('eav_widget', $eav_widget);

    return $form;
  }

  /**
   * {@inheritDoc}
   */
  public function submitForm(array &$form, FormStateInterface $form_state) {
    $entity = $form_state->get('entity'); /** @var EntityInterface $entity */
    $eav_widget = $form_state->get('eav_widget'); /** @var WidgetInterface $eav_widget */
    $eav_field_name = $this->getEntityEavFieldDefinition($entity)->getName();

    $eav_widget->extractFormValues($entity->get($eav_field_name), $form, $form_state);
    $entity->save();

    $this->messenger()->addMessage($this->t('Values saved.'));
  }

  /**
   * "delete" button submit callback.
   */
  public function clearAllValues(array &$form, FormStateInterface $form_state) {
    $entity = $form_state->get('entity'); /** @var EntityInterface $entity */
    $eav_field_name = $this->getEntityEavFieldDefinition($entity)->getName();

    // Delete all eav_value entities and clear eav field value
    $entity->set($eav_field_name, NULL);
    $entity->save();

    $this->messenger()->addMessage($this->t('Values cleared.'));
  }

  /**
   * Return entity from route.
   */
  protected function getEntityFromRoute(): ?EntityInterface {
    $query_parameters = $this->getRouteMatch()->getRawParameters()->all();
    $entity_type_id = key($query_parameters);
    $entity_id = current($query_parameters);
    return $this->entityTypeManager->getStorage($entity_type_id)->load($entity_id);
  }

  /**
   * Return eav field definition.
   */
  protected function getEntityEavFieldDefinition(EntityInterface $entity): ?FieldDefinitionInterface {
    $entity_bundle_field_definitions = $this->entityFieldManager->getFieldDefinitions($entity->getEntityTypeId(), $entity->bundle());
    $eav_field_definition = NULL;

    foreach ($entity_bundle_field_definitions as $entity_bundle_field_definition) {
      if ($entity_bundle_field_definition->getType() == 'eav') {
        $eav_field_definition = $entity_bundle_field_definition;
        break;
      }
    }

    return $eav_field_definition;
  }

}
