<?php

namespace Drupal\eav_field\Plugin\Field\FieldWidget;

use Drupal\Core\Entity\ContentEntityInterface;
use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\Core\Field\Annotation\FieldWidget;
use Drupal\Core\Field\FieldDefinitionInterface;
use Drupal\Core\Field\FieldItemListInterface;
use Drupal\Core\Field\WidgetBase;
use Drupal\Core\Field\WidgetPluginManager;
use Drupal\Core\Form\FormStateInterface;
use Drupal\Core\Plugin\ContainerFactoryPluginInterface;
use Drupal\Core\Render\Element;
use Drupal\eav_field\EavAttributeStorage;
use Drupal\eav_field\Entity\EavAttribute;
use Drupal\eav_field\Entity\EavValue;
use Drupal\eav_field\Plugin\Field\FieldType\EavItem;
use Drupal\eav_field\Plugin\Field\FieldType\EavItemList;
use Drupal\field\Entity\FieldConfig;
use Symfony\Component\DependencyInjection\ContainerInterface;

/**
 * @FieldWidget(
 *   id = "eav_widget",
 *   label = @Translation("EAV widget"),
 *   field_types = {"eav"},
 *   multiple_values = TRUE,
 * )
 */
class EavWidget extends WidgetBase {

  /**
   * @var EntityTypeManagerInterface
   */
  protected $entityTypeManager;

  /**
   * @var WidgetPluginManager
   */
  protected $fieldWidgetManager;

  /**
   * {@inheritdoc}
   */
  public function __construct($plugin_id, $plugin_definition, FieldDefinitionInterface $field_definition, array $settings, array $third_party_settings, EntityTypeManagerInterface $entity_type_manager, WidgetPluginManager $field_widget_manager) {
    parent::__construct($plugin_id, $plugin_definition, $field_definition, $settings, $third_party_settings);
    $this->entityTypeManager = $entity_type_manager;
    $this->fieldWidgetManager = $field_widget_manager;
  }

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container, array $configuration, $plugin_id, $plugin_definition) {
    return new static(
      $plugin_id,
      $plugin_definition,
      $configuration['field_definition'],
      $configuration['settings'],
      $configuration['third_party_settings'],
      $container->get('entity_type.manager'),
      $container->get('plugin.manager.field.widget')
    );
  }

  /**
   * {@inheritdoc}
   */
  public function formElement(FieldItemListInterface $items, $delta, array $element, array &$form, FormStateInterface $form_state) {
    /** @var EavItemList $items */

    $element['table'] = [
      '#type' => 'table',
      '#header' => [
        $this->t('Attribute'),
        $this->t('Value'),
        '', // For "value_id" field
      ],
      '#parents' => array_merge($element['#field_parents'], [$items->getName()]),
      '#empty' => $this->t('No available attributes.'),
    ];

    $element['target_ids']['#tree'] = TRUE;

    if ($attributes = $this->getAttributesByCurrentCategory($items)) {
      $eav_values = [];
      foreach ($items as $item) {
        /** @var EavItem $item */
        if (($eav_value = $item->getValueEntity()) && ($attribute = $eav_value->getAttributeEntity())) {
          $eav_values[$attribute->getMachineName()] = $item->getValueEntity();
        }
      }

      foreach ($attributes as $attribute) {
        // Value field need configure before EavValue::create() and before $eav_value->getValueFieldItems()
        $attribute->configureValueFieldDefinition();

        $attribute_machine_name = $attribute->getMachineName();
        $eav_value = $eav_values[$attribute_machine_name] ?? EavValue::create(['aid' => $attribute->id()]);
        $value_field_name = $attribute->getValueFieldName();
        $value_field_items = $eav_value->getValueFieldItems();

        $element['table'][$attribute_machine_name]['#eav_attribute'] = $attribute;

        $element['table'][$attribute_machine_name]['attribute'] = [
          '#markup' => $attribute->label(),
        ];

        $element['table'][$attribute_machine_name][$value_field_name] = $this->valueWidget($attribute, $value_field_items, $element, $form, $form_state);

        // @TODO Find way of pass value_id in ::massageFormValues without create table cell
        $element['table'][$attribute_machine_name]['value_id'] = [
          '#type' => 'value',
          '#value' => $eav_value->id(),
        ];
      }
    }
    else {
      $element['table']['#value'] = [];
    }

    return $element;
  }

  /**
   * Return attributes by current category.
   *
   * @return EavAttribute[]
   */
  protected function getAttributesByCurrentCategory(FieldItemListInterface $items): array {
    $attribute_storage = $this->entityTypeManager->getStorage('eav_attribute'); /** @var EavAttributeStorage $attribute_storage */
    $attributes = $attribute_storage->loadByCategory(NULL); // Load global attributes (without categories)
    $host_entity = $items->getEntity();
    if ($host_entity_categories = $this->getHostEntityCategories($host_entity)) {
      $attributes += $attribute_storage->loadByCategory($host_entity_categories); // Load attributes with certain category (including parents categories)
    }

    return $attributes;
  }

  /**
   * Return host entity category field definition.
   *
   * @return FieldDefinitionInterface|NULL
   */
  protected function getHostEntityCategoryField(ContentEntityInterface $host_entity) {
    $attribute_category_field = FieldConfig::loadByName('eav_attribute', 'eav_attribute', 'category');
    $attribute_category_field_storage = $attribute_category_field->getFieldStorageDefinition();
    $attribute_category_field_handler_settings = $attribute_category_field->getSetting('handler_settings');

    if ($attribute_category_field_handler_settings) {
      $attribute_category_target_type = $attribute_category_field_storage->getSetting('target_type');
      $attribute_category_target_bundle = key($attribute_category_field_handler_settings['target_bundles']);

      foreach ($host_entity->getFieldDefinitions() as $host_entity_field_name => $host_entity_field) {
        if (
          $host_entity_field->getType() == 'entity_reference' &&
          $host_entity_field->getFieldStorageDefinition()->getSetting('target_type') == $attribute_category_target_type &&
          isset($host_entity_field->getSetting('handler_settings')['target_bundles'][$attribute_category_target_bundle])
        ) {
          return $host_entity_field;
        }
      }
    }

    return NULL;
  }

  /**
   * Return host entity category.
   *
   * @return array
   */
  protected function getHostEntityCategories(ContentEntityInterface $host_entity) {
    if ($host_entity_category_field = $this->getHostEntityCategoryField($host_entity)) {
      $host_entity_categories = [];

      foreach ($host_entity->get($host_entity_category_field->getName()) as $item) {
        $host_entity_categories[] = $item->target_id;
      }

      return $host_entity_categories;
    }

    return NULL;
  }

  /**
   * Return value field widget.
   */
  protected function valueWidget(EavAttribute $attribute, FieldItemListInterface $value_field_items, array $element, array $form, FormStateInterface $form_state) {
    $widget = $this->fieldWidgetManager->getInstance([
      'field_definition' => $attribute->getValueFieldDefinition(),
      'configuration' => [
        'type' => $attribute->getValueWidgetType(),
        'settings' => $attribute->getValueWidgetSettings(),
      ],
    ]);

    $form['#parents'] = array_merge($element['table']['#parents'], [$attribute->getMachineName()]);

    // @TODO Use $element without $form
    $widget_form = $widget->form($value_field_items, $form, $form_state);

    // Hide widget title
    $this->hideElementTitle($widget_form['widget']);
    foreach (Element::children($widget_form['widget']) as $key) {
      $this->hideElementTitle($widget_form['widget'][$key]);
      if (isset($widget_form['widget'][$key]['value'])) {
        $this->hideElementTitle($widget_form['widget'][$key]['value']);
      }
    }

    return $widget_form;
  }

  /**
   * Hide element title.
   */
  protected function hideElementTitle(&$element) {
    if (isset($element['#title'])) {
      $element['#title_display'] = 'invisible';
    }
  }

  /**
   * {@inheritdoc}
   */
  public function massageFormValues(array $values, array $form, FormStateInterface $form_state) {
    $new_values = [];

    foreach ($values as $attribute_name => $item_values) {
      $attribute = EavAttribute::loadByMachineName($attribute_name);
      /** @var EavValue $eav_value */
      $eav_value = $item_values['value_id']
        ? EavValue::load($item_values['value_id'])
        : EavValue::create(['aid' => $attribute->id()]);

      $value_field_widget = $this->fieldWidgetManager->getInstance([
        'field_definition' => $attribute->configureValueFieldDefinition(),
        'configuration' => [
          'type' => $attribute->getValueWidgetType(),
          'settings' => $attribute->getValueWidgetSettings(),
        ],
      ]);

      $value_field_items = $eav_value->getValueFieldItems();
      $value_field_form = $form[$this->fieldDefinition->getName()]['widget']['table'][$attribute_name];
      $value_field_widget->extractFormValues($value_field_items, $value_field_form, $form_state);

      $new_values[]['entity'] = $eav_value;
    }

    return $new_values;
  }

}
