<?php

namespace Drupal\Tests\eav_field\Functional;

use Drupal\commerce_product\Entity\Product;
use Drupal\commerce_store\Entity\Store;
use Drupal\Core\Url;
use Drupal\eav_field\Entity\EavAttribute;
use Drupal\eav_field\Entity\EavValue;
use Drupal\field\Entity\FieldConfig;
use Drupal\node\Entity\Node;
use Drupal\taxonomy\Entity\Vocabulary;
use Drupal\taxonomy\VocabularyInterface;
use Drupal\Tests\block\Traits\BlockCreationTrait;
use Drupal\Tests\BrowserTestBase;
use Drupal\Tests\eav_field\Traits\EavFieldTestTrait;
use Drupal\Tests\improvements\Traits\ImprovementsTestTrait;
use Drupal\Tests\node\Traits\NodeCreationTrait;
use Drupal\Tests\taxonomy\Traits\TaxonomyTestTrait;

class EavFieldTest extends BrowserTestBase {

  use BlockCreationTrait;
  use EavFieldTestTrait;
  use ImprovementsTestTrait;
  use NodeCreationTrait;
  use TaxonomyTestTrait;

  /**
   * {@inheritDoc}
   */
  protected $defaultTheme = 'classy';

  /**
   * {@inheritDoc}
   */
  protected static $modules = [
    'block',
    'eav_field',
    'field_ui',
    'node',
    'taxonomy',
    'user',
  ];

  /**
   * @var VocabularyInterface
   */
  protected $vocabulary;

  protected function setUp() {
    parent::setUp();

    // Create "product" node type
    $this->createContentType(['type' => 'product', 'name' => 'Product']);

    // Create "category" vocabulary
    $this->vocabulary = Vocabulary::create(['name' => 'Categories', 'vid' => 'category']);
    $this->vocabulary->save();

    // Create categories
    $this->createTerm($this->vocabulary, ['name' => 'Category 1']); // 1
    $this->createTerm($this->vocabulary, ['name' => 'Category 2']); // 2
    $this->createTerm($this->vocabulary, ['name' => 'Category 3']); // 3
    $this->createTerm($this->vocabulary, ['name' => 'Category 1.1', 'parent' => 1]); // 4
    $this->createTerm($this->vocabulary, ['name' => 'Category 1.2', 'parent' => 1]); // 5

    // Create "field_category" node field
    $this->createField(
      'node',
      'product',
      'field_category',
      'entity_reference',
      'Category',
      ['settings' => [
        'target_type' => 'taxonomy_term'
      ]],
      ['settings' => [
        'handler' => 'default:taxonomy_term',
        'handler_settings' => [
          'target_bundles' => [
            'category' => 'category',
          ],
        ],
      ]],
      'options_select',
      [],
      'entity_reference_label',
      []
    );

    // Module initial setup
    $eav_attribute_category_field = FieldConfig::load('eav_attribute.eav_attribute.category');
    $eav_attribute_category_field_settings = $eav_attribute_category_field->get('settings');
    $eav_attribute_category_field_settings['handler_settings']['target_bundles']['category'] = 'category';
    $eav_attribute_category_field->set('settings', $eav_attribute_category_field_settings);
    $eav_attribute_category_field->save();

    // Create field "field_eav"
    $this->createField(
      'node',
      'product',
      'field_eav',
      'eav',
      'EAV attributes',
      [],
      [],
      'eav_widget',
      [],
      'eav_formatter',
      []
    );
    drupal_flush_all_caches();

    // Create product nodes
    $this->createNode([
      'title' => 'Product 1',
      'type' => 'product',
      'field_category' => 1,
    ]);
    $this->createNode([
      'title' => 'Product 2',
      'type' => 'product',
      'field_category' => 1,
    ]);

    // Place blocks
    $this->placeBlock('page_title_block', [
      'id' => 'page_title_block',
      'region' => 'content',
      'weight' => -100,
    ]);
    $this->placeBlock('local_tasks_block', [
      'id' => 'local_tasks_block',
      'region' => 'content',
      'weight' => -99,
    ]);
    $this->placeBlock('local_actions_block', [
      'id' => 'local_actions_block',
      'region' => 'content',
      'weight' => -98,
    ]);
  }

  public function testMain() {
    // Run all test with "_test" prefix
    $class_methods = get_class_methods(self::class);
    foreach ($class_methods as $class_method) {
      if (strpos($class_method, '_test') === 0) {
        $result = $this->{$class_method}();
        if ($result === FALSE) {
          break;
        }
      }
    }
  }

  private function _testInitialSetup() {
    $web_assert = $this->assertSession();
    $this->drupalLoginAsRoot();

    $this->drupalGet('admin/structure/eav/attributes/fields/eav_attribute.eav_attribute.category');
    $this->dontSeeErrorMessage();
    $web_assert->fieldExists('settings[handler_settings][target_bundles][category]');
    $web_assert->checkboxChecked('settings[handler_settings][target_bundles][category]');

    $this->drupalGet('admin/structure/types/manage/product/fields');
    $this->dontSeeErrorMessage();
    $web_assert->elementExists('css', '#field-eav');
    $web_assert->elementExists('css', '#field-category');

    $this->drupalGet('admin/structure/types/manage/product/form-display');
    $this->dontSeeErrorMessage();
    $web_assert->fieldValueEquals('fields[field_eav][type]', 'eav_widget');
    $web_assert->fieldValueEquals('fields[field_category][type]', 'options_select');

    $this->drupalGet('admin/structure/types/manage/product/display');
    $this->dontSeeErrorMessage();
    $web_assert->fieldValueEquals('fields[field_eav][type]', 'eav_formatter');
    $web_assert->fieldValueEquals('fields[field_category][type]', 'entity_reference_label');
  }

  private function _testAddEavFieldFromUi() {
    $web_assert = $this->assertSession();
    $this->drupalLoginAsRoot();

    $this->drupalGet('admin/structure/types/manage/product/fields/add-field');
    $this->dontSeeErrorMessage();
    $web_assert->optionExists('new_storage_type', 'eav');

    $this->submitForm([
      'new_storage_type' => 'eav',
      'label' => 'Test EAV attributes field',
      'field_name' => 'test_eav_attributes',
    ], 'Save and continue');
    $this->dontSeeErrorMessage();
    $web_assert->elementContains('css', 'h1', 'Test EAV attributes field');

    $this->submitForm([], 'Save field settings');
    $this->dontSeeErrorMessage();
    $web_assert->fieldValueEquals('label', 'Test EAV attributes field');

    $this->submitForm([], 'Save settings');
    $this->dontSeeErrorMessage();
    $web_assert->pageTextContains('Saved Test EAV attributes field configuration.');
    $web_assert->pageTextContains('field_test_eav_attributes');

    $this->drupalGet('admin/structure/types/manage/product/form-display');
    $this->dontSeeErrorMessage();
    $web_assert->fieldValueEquals('fields[field_eav][type]', 'eav_widget');

    $this->drupalGet('admin/structure/types/manage/product/display');
    $this->dontSeeErrorMessage();
    $web_assert->fieldValueEquals('fields[field_eav][type]', 'eav_formatter');

    $this->drupalGet('admin/structure/types/manage/product/fields/node.product.field_test_eav_attributes/delete');
    $this->dontSeeErrorMessage();
    $this->submitForm([], 'Delete');
    $this->dontSeeErrorMessage();
    $web_assert->pageTextMatches('/The field .+ has been deleted from the .+ content type/');
    $web_assert->pageTextNotContains('field_test_eav_attributes');
  }

  private function _testAddStringAttributeFromUi() {
    $web_assert = $this->assertSession();
    $this->drupalLoginAsRoot();


    $this->drupalGet(Url::fromRoute('entity.eav_attribute.collection'));
    $this->dontSeeErrorMessage();
    $web_assert->elementContains('css', 'h1', 'Attributes');
    $web_assert->elementExists('css', 'ul.primary a[href="/admin/structure/eav/attributes/fields"]');
    $web_assert->elementExists('css', 'ul.primary a[href="/admin/structure/eav/attributes/form-display"]');
    $web_assert->elementExists('css', 'ul.primary a[href="/admin/structure/eav/attributes/display"]');
    $web_assert->elementExists('css', 'form#eav-attribute-list');

    $attribute_values = [
      'label' => 'String attribute',
      'machine_name' => 'string_attr',
      'value_type' => 'string',
      'value_storage_config' => [
        'cardinality' => 1,
        'settings' => [
          'max_length' => 200,
        ],
      ],
      'value_config' => [
        'description' => 'String attribute value field description',
        'required' => FALSE,
        'default_value' => [
          0 => [
            'value' => 'String attribute value field default value',
          ],
        ],
        'settings' => [],
      ],
      'value_widget_type' => 'string_textfield',
      'value_widget_settings' => [
        'size' => 50,
        'placeholder' => 'String attribute value field placeholder',
      ],
      'value_formatter_type' => 'string',
      'value_formatter_settings' => [
        'link_to_entity' => FALSE,
      ],
      'description' => '',
      'category' => [],
    ];

    // Attribute "add" form
    $this->clickLink('Add attribute');
    $this->dontSeeErrorMessage();
    $web_assert->elementContains('css', 'h1', 'Adding attribute');
    $web_assert->fieldExists('label[0][value]');
    $web_assert->fieldExists('admin_label[0][value]');
    $web_assert->fieldExists('machine_name[0][value]');
    $web_assert->fieldExists('value_type');
    $web_assert->fieldExists('value_widget_type');
    $web_assert->fieldExists('value_formatter_type');
    $web_assert->fieldExists('category[0][target_id]');
    $web_assert->fieldExists('description[0][value]');
    $web_assert->fieldExists('primary[value]');
    $web_assert->fieldExists('weight[0][value]');

    $form_fields_values = [
      'label[0][value]' => $attribute_values['label'],
      'machine_name[0][value]' => $attribute_values['machine_name'],
      'value_type' => $attribute_values['value_type'],
      'value_widget_type' => $attribute_values['value_widget_type'],
      'value_formatter_type' => $attribute_values['value_formatter_type'],
      'category[0][target_id]' => '',
      'description[0][value]' => $attribute_values['value_config']['default_value'][0]['value'],
    ];
    $this->submitForm($form_fields_values, 'Save');
    $this->dontSeeErrorMessage();

    $attribute_id = $this->getLastAddedEavAttributeId();
    $this->drupalGet(Url::fromRoute('entity.eav_attribute.edit_form', ['eav_attribute' => $attribute_id]));
    $this->dontSeeErrorMessage();
    foreach ($form_fields_values as $field_name => $field_value) {
      $web_assert->fieldValueEquals($field_name, $field_value);
    }

    // Storage settings form
    $this->drupalGet(Url::fromRoute('entity.eav_attribute.value_storage_form', ['eav_attribute' => $attribute_id]));
    $this->dontSeeErrorMessage();
    $web_assert->pageTextContains('Storage settings(active tab)');
    $web_assert->fieldExists('settings[max_length]');
    $web_assert->fieldEnabled('settings[max_length]');
    $web_assert->fieldExists('cardinality');
    $web_assert->fieldExists('cardinality_number');
    $this->submitForm([
      'settings[max_length]' => $attribute_values['value_storage_config']['settings']['max_length'],
    ], 'Save field settings');
    $this->dontSeeErrorMessage();
    $web_assert->fieldValueEquals('settings[max_length]', $attribute_values['value_storage_config']['settings']['max_length']);
    $web_assert->fieldEnabled('settings[max_length]');
    $web_assert->fieldValueEquals('cardinality', 'number');
    $web_assert->fieldValueEquals('cardinality_number', $attribute_values['value_storage_config']['cardinality']);

    // Field settings form
    $this->drupalGet(Url::fromRoute('entity.eav_attribute.value_field_form', ['eav_attribute' => $attribute_id]));
    $this->dontSeeErrorMessage();
    $web_assert->pageTextContains('Field settings(active tab)');
    $web_assert->fieldExists('description');
    $web_assert->fieldExists('required');
    $web_assert->fieldExists('default_value_input[string_value][0][value]');
    $this->submitForm([
      'description' => $attribute_values['value_config']['description'],
      'default_value_input[string_value][0][value]' => $attribute_values['value_config']['default_value'][0]['value'],
    ], 'Save settings');
    $this->dontSeeErrorMessage();
    $web_assert->fieldValueEquals('description', $attribute_values['value_config']['description']);
    $web_assert->fieldValueEquals('default_value_input[string_value][0][value]', $attribute_values['value_config']['default_value'][0]['value']);

    // Widget settings form
    $this->drupalGet(Url::fromRoute('entity.eav_attribute.value_widget_form', ['eav_attribute' => $attribute_id]));
    $this->dontSeeErrorMessage();
    $web_assert->pageTextContains('Widget settings(active tab)');
    $web_assert->fieldExists('fields[string_value][settings_edit_form][settings][size]');
    $web_assert->fieldExists('fields[string_value][settings_edit_form][settings][placeholder]');
    $this->submitForm([
      'fields[string_value][settings_edit_form][settings][size]' => $attribute_values['value_widget_settings']['size'],
      'fields[string_value][settings_edit_form][settings][placeholder]' => $attribute_values['value_widget_settings']['placeholder'],
    ], 'Save');
    $this->dontSeeErrorMessage();
    $web_assert->fieldValueEquals('fields[string_value][settings_edit_form][settings][size]', $attribute_values['value_widget_settings']['size']);
    $web_assert->fieldValueEquals('fields[string_value][settings_edit_form][settings][placeholder]', $attribute_values['value_widget_settings']['placeholder']);

    // Formatter settings form
    $this->drupalGet(Url::fromRoute('entity.eav_attribute.value_formatter_form', ['eav_attribute' => $attribute_id]));
    $this->dontSeeErrorMessage();
    $web_assert->pageTextContains('Formatter settings(active tab)');
    $web_assert->fieldExists('fields[string_value][settings_edit_form][settings][link_to_entity]');
    $this->submitForm([], 'Save');
    $this->dontSeeErrorMessage();

    // Attributes list
    $this->drupalGet(Url::fromRoute('entity.eav_attribute.collection'));
    $this->dontSeeErrorMessage();
    $attribute_tr_selector = 'tr[data-drupal-selector="edit-eav-attributes-' . $attribute_id . '"]';
    $web_assert->elementExists('css', $attribute_tr_selector);
    $web_assert->elementExists('css', $attribute_tr_selector . ' ul.dropbutton li.edit');
    $web_assert->elementExists('css', $attribute_tr_selector . ' ul.dropbutton li.edit-storage');
    $web_assert->elementExists('css', $attribute_tr_selector . ' ul.dropbutton li.edit-field');
    $web_assert->elementExists('css', $attribute_tr_selector . ' ul.dropbutton li.edit-widget');
    $web_assert->elementExists('css', $attribute_tr_selector . ' ul.dropbutton li.edit-formatter');
    $web_assert->elementExists('css', $attribute_tr_selector . ' ul.dropbutton li.delete');
  }

  private function _testAddSecondStringAttributeFromCode() {
    $web_assert = $this->assertSession();
    $this->drupalLoginAsRoot();

    $attribute_values = [
      'label' => 'Second string attribute',
      'machine_name' => 'second_string_attr',
      'value_type' => 'string',
      'value_storage_config' => [
        'cardinality' => 1,
        'settings' => [
          'max_length' => 210,
        ],
      ],
      'value_config' => [
        'description' => '',
        'required' => FALSE,
        'default_value' => [],
        'settings' => [],
      ],
      'value_widget_type' => 'string_textfield',
      'value_widget_settings' => [
        'size' => 40,
        'placeholder' => '',
      ],
      'value_formatter_type' => 'string',
      'value_formatter_settings' => [
        'link_to_entity' => FALSE,
      ],
      'description' => '',
      'category' => [1],
    ];
    $attribute = EavAttribute::create($attribute_values); /** @var EavAttribute $attribute */
    $attribute->save();

    // Attributes list
    $this->drupalGet(Url::fromRoute('entity.eav_attribute.collection'));
    $this->dontSeeErrorMessage();
    $web_assert->elementExists('css', 'tr[data-drupal-selector="edit-eav-attributes-' . $attribute->id() . '"]');

    // Attribute edit form
    $this->drupalGet('admin/structure/eav/attributes/' . $attribute->id() . '/edit');
    $this->dontSeeErrorMessage();
    $web_assert->fieldValueEquals('label[0][value]', $attribute_values['label']);
    $web_assert->fieldValueEquals('machine_name[0][value]', $attribute_values['machine_name']);
    $web_assert->fieldValueEquals('value_type', $attribute_values['value_type']);
    $web_assert->fieldValueEquals('value_widget_type', $attribute_values['value_widget_type']);
    $web_assert->fieldValueEquals('value_formatter_type', $attribute_values['value_formatter_type']);
    $web_assert->fieldValueEquals('category[0][target_id]', 'Category 1 (1)');
    $web_assert->fieldValueEquals('description[0][value]', $attribute_values['description']);

    // Storage settings form
    $this->drupalGet('admin/structure/eav/attributes/' . $attribute->id() . '/storage');
    $this->dontSeeErrorMessage();
    $web_assert->fieldValueEquals('settings[max_length]', $attribute_values['value_storage_config']['settings']['max_length']);
    $web_assert->fieldValueEquals('cardinality', 'number');
    $web_assert->fieldValueEquals('cardinality_number', $attribute_values['value_storage_config']['cardinality']);

    // Field settings form
    $this->drupalGet('admin/structure/eav/attributes/' . $attribute->id() . '/field');
    $this->dontSeeErrorMessage();
    $web_assert->fieldValueEquals('description', $attribute_values['value_config']['description']);
    $web_assert->fieldValueEquals('default_value_input[string_value][0][value]', '');

    // Widget settings form
    $this->drupalGet('admin/structure/eav/attributes/' . $attribute->id() . '/widget');
    $this->dontSeeErrorMessage();
    $web_assert->fieldValueEquals('fields[string_value][settings_edit_form][settings][size]', $attribute_values['value_widget_settings']['size']);
    $web_assert->fieldValueEquals('fields[string_value][settings_edit_form][settings][placeholder]', $attribute_values['value_widget_settings']['placeholder']);

    // Formatter form
    $this->drupalGet('admin/structure/eav/attributes/' . $attribute->id() . '/formatter');
    $this->dontSeeErrorMessage();
    $web_assert->checkboxNotChecked('fields[string_value][settings_edit_form][settings][link_to_entity]');
  }

  private function _testAddStringLongAttributeFromCode() {
    $web_assert = $this->assertSession();
    $this->drupalLoginAsRoot();

    $attribute_values = [
      'label' => 'Long string attribute',
      'machine_name' => 'string_long_attr',
      'value_type' => 'string_long',
      'value_storage_config' => [
        'cardinality' => 1,
        'settings' => [
          'case_sensitive' => FALSE,
        ],
      ],
      'value_config' => [
        'description' => 'Value field description',
        'required' => FALSE,
        'default_value' => [],
        'settings' => [],
      ],
      'value_widget_type' => 'string_textarea',
      'value_widget_settings' => [
        'rows' => 6,
        'placeholder' => 'Long string attribute value placeholder',
      ],
      'value_formatter_type' => 'basic_string',
      'value_formatter_settings' => [],
      'description' => 'Attribute description',
      'category' => [1],
    ];
    $attribute = EavAttribute::create($attribute_values); /** @var EavAttribute $attribute */
    $attribute->save();

    // Attributes list
    $this->drupalGet('admin/structure/eav/attributes');
    $this->dontSeeErrorMessage();
    $web_assert->elementExists('css', 'tr[data-drupal-selector="edit-eav-attributes-' . $attribute->id() . '"]');

    // Attribute edit form
    $this->drupalGet('admin/structure/eav/attributes/' . $attribute->id() . '/edit');
    $this->dontSeeErrorMessage();
    $web_assert->fieldValueEquals('label[0][value]', $attribute->label());
    $web_assert->fieldValueEquals('machine_name[0][value]', $attribute->getMachineName());
    $web_assert->fieldValueEquals('value_type', $attribute->getValueType());
    $web_assert->fieldValueEquals('value_widget_type', $attribute->getValueWidgetType());
    $web_assert->fieldValueEquals('value_formatter_type', $attribute->getValueFormatterType());
    $web_assert->fieldValueEquals('category[0][target_id]', 'Category 1 (1)');
    $web_assert->fieldValueEquals('description[0][value]', $attribute->get('description')->value);

    // Storage settings form
    $this->drupalGet('admin/structure/eav/attributes/' . $attribute->id() . '/storage');
    $this->dontSeeErrorMessage();
    $web_assert->fieldValueEquals('cardinality', 'number');
    $web_assert->fieldValueEquals('cardinality_number', '1');

    // Field settings form
    $this->drupalGet('admin/structure/eav/attributes/' . $attribute->id() . '/field');
    $this->dontSeeErrorMessage();
    $web_assert->fieldValueEquals('description', $attribute_values['value_config']['description']);
    $web_assert->fieldValueEquals('default_value_input[string_long_value][0][value]', '');

    // Widget settings form
    $this->drupalGet('admin/structure/eav/attributes/' . $attribute->id() . '/widget');
    $this->dontSeeErrorMessage();
    $web_assert->fieldValueEquals('fields[string_long_value][settings_edit_form][settings][rows]', $attribute_values['value_widget_settings']['rows']);
    $web_assert->fieldValueEquals('fields[string_long_value][settings_edit_form][settings][placeholder]', $attribute_values['value_widget_settings']['placeholder']);

    // Formatter form
    $this->drupalGet('admin/structure/eav/attributes/' . $attribute->id() . '/formatter');
    $this->dontSeeErrorMessage();
    $web_assert->pageTextContains('No settings.');
  }

  private function _testAddIntegerAttributeFromCode() {
    $web_assert = $this->assertSession();
    $this->drupalLoginAsRoot();

    $attribute_values = [
      'label' => 'Integer attribute',
      'machine_name' => 'integer_attr',
      'value_type' => 'integer',
      'value_storage_config' => [
        'cardinality' => 1,
        'settings' => [
          'unsigned' => FALSE,
          'size' => 'normal',
        ],
      ],
      'value_config' => [
        'description' => '',
        'required' => FALSE,
        'default_value' => [],
        'settings' => [
          'min' => '0',
          'max' => '1000',
          'prefix' => '',
          'suffix' => 'sffx',
        ],
      ],
      'value_widget_type' => 'number',
      'value_widget_settings' => [
        'placeholder' => '',
      ],
      'value_formatter_type' => 'number_integer',
      'value_formatter_settings' => [
        'thousand_separator' => ' ',
        'prefix_suffix' => TRUE,
      ],
      'description' => '',
      'category' => [1],
    ];
    $attribute = EavAttribute::create($attribute_values);
    $attribute->save();

    // Attributes list
    $this->drupalGet('admin/structure/eav/attributes');
    $this->dontSeeErrorMessage();
    $web_assert->elementExists('css', 'tr[data-drupal-selector="edit-eav-attributes-' . $attribute->id() . '"]');

    // Attribute form
    $this->drupalGet('admin/structure/eav/attributes/' . $attribute->id() . '/edit');
    $this->dontSeeErrorMessage();
    $web_assert->fieldValueEquals('label[0][value]', $attribute->label());
    $web_assert->fieldValueEquals('machine_name[0][value]', $attribute->getMachineName());
    $web_assert->fieldValueEquals('value_type', $attribute->getValueType());
    $web_assert->fieldValueEquals('value_widget_type', $attribute->getValueWidgetType());
    $web_assert->fieldValueEquals('value_formatter_type', $attribute->getValueFormatterType());
    $web_assert->fieldValueEquals('category[0][target_id]', 'Category 1 (1)');
    $web_assert->fieldValueEquals('description[0][value]', $attribute->get('description')->value);

    // Storage settings form
    $this->drupalGet('admin/structure/eav/attributes/' . $attribute->id() . '/storage');
    $this->dontSeeErrorMessage();
    $web_assert->fieldValueEquals('cardinality', 'number');
    $web_assert->fieldValueEquals('cardinality_number', '1');

    // Field settings form
    $this->drupalGet('admin/structure/eav/attributes/' . $attribute->id() . '/field');
    $this->dontSeeErrorMessage();
    $web_assert->fieldValueEquals('settings[min]', $attribute_values['value_config']['settings']['min']);
    $web_assert->fieldValueEquals('settings[max]', $attribute_values['value_config']['settings']['max']);
    $web_assert->fieldValueEquals('settings[prefix]', $attribute_values['value_config']['settings']['prefix']);
    $web_assert->fieldValueEquals('settings[suffix]', $attribute_values['value_config']['settings']['suffix']);
    $web_assert->fieldValueEquals('default_value_input[integer_value][0][value]', '');

    // Widget settings form
    $this->drupalGet('admin/structure/eav/attributes/' . $attribute->id() . '/widget');
    $this->dontSeeErrorMessage();
    $web_assert->fieldValueEquals('fields[integer_value][settings_edit_form][settings][placeholder]', '');

    // Formatter settings form
    $this->drupalGet('admin/structure/eav/attributes/' . $attribute->id() . '/formatter');
    $this->dontSeeErrorMessage();
    $web_assert->fieldValueEquals('fields[integer_value][settings_edit_form][settings][thousand_separator]', ' ');
    $web_assert->checkboxChecked('fields[integer_value][settings_edit_form][settings][prefix_suffix]');
  }

  private function _testAddDecimalAttributeFromCode() {
    $web_assert = $this->assertSession();
    $this->drupalLoginAsRoot();

    $attribute_values = [
      'label' => 'Decimal attribute',
      'machine_name' => 'decimal_attr',
      'value_type' => 'decimal',
      'value_storage_config' => [
        'cardinality' => 1,
        'settings' => [
          'precision' => 10,
          'scale' => 2,
        ],
      ],
      'value_config' => [
        'description' => '',
        'required' => FALSE,
        'default_value' => [],
        'settings' => [
          'min' => '-10',
          'max' => '300',
          'prefix' => '$',
          'suffix' => '',
        ],
      ],
      'value_widget_type' => 'number',
      'value_widget_settings' => [
        'placeholder' => '',
      ],
      'value_formatter_type' => 'number_decimal',
      'value_formatter_settings' => [
        'thousand_separator' => ' ',
        'decimal_separator' => '.',
        'scale' => 2,
        'prefix_suffix' => TRUE,
      ],
      'description' => '',
      'category' => [1],
    ];
    $attribute = EavAttribute::create($attribute_values); /** @var EavAttribute $attribute */
    $attribute->save();

    // Attributes list
    $this->drupalGet('admin/structure/eav/attributes');
    $this->dontSeeErrorMessage();
    $web_assert->elementExists('css', 'tr[data-drupal-selector="edit-eav-attributes-' . $attribute->id() . '"]');

    // Attribute form
    $this->drupalGet('admin/structure/eav/attributes/' . $attribute->id() . '/edit');
    $this->dontSeeErrorMessage();
    $web_assert->fieldValueEquals('label[0][value]', $attribute_values['label']);
    $web_assert->fieldValueEquals('machine_name[0][value]', $attribute_values['machine_name']);
    $web_assert->fieldValueEquals('value_type', $attribute_values['value_type']);
    $web_assert->fieldValueEquals('value_widget_type', $attribute_values['value_widget_type']);
    $web_assert->fieldValueEquals('value_formatter_type', $attribute_values['value_formatter_type']);
    $web_assert->fieldValueEquals('category[0][target_id]', 'Category 1 (1)');
    $web_assert->fieldValueEquals('description[0][value]', $attribute_values['description']);

    // Storage settings form
    $this->drupalGet('admin/structure/eav/attributes/' . $attribute->id() . '/storage');
    $this->dontSeeErrorMessage();
    $web_assert->fieldValueEquals('cardinality', 'number');
    $web_assert->fieldValueEquals('cardinality_number', $attribute_values['value_storage_config']['cardinality']);
    $web_assert->fieldValueEquals('settings[precision]', $attribute_values['value_storage_config']['settings']['precision']);
    $web_assert->fieldValueEquals('settings[scale]', $attribute_values['value_storage_config']['settings']['scale']);

    // Field settings form
    $this->drupalGet('admin/structure/eav/attributes/' . $attribute->id() . '/field');
    $this->dontSeeErrorMessage();
    $web_assert->fieldValueEquals('settings[min]', $attribute_values['value_config']['settings']['min']);
    $web_assert->fieldValueEquals('settings[max]', $attribute_values['value_config']['settings']['max']);
    $web_assert->fieldValueEquals('settings[prefix]', $attribute_values['value_config']['settings']['prefix']);
    $web_assert->fieldValueEquals('settings[suffix]', $attribute_values['value_config']['settings']['suffix']);
    $web_assert->fieldValueEquals('default_value_input[decimal_value][0][value]', '');

    // Widget settings form
    $this->drupalGet('admin/structure/eav/attributes/' . $attribute->id() . '/widget');
    $this->dontSeeErrorMessage();
    $web_assert->fieldValueEquals('fields[decimal_value][settings_edit_form][settings][placeholder]', $attribute_values['value_widget_settings']['placeholder']);

    // Formatter settings form
    $this->drupalGet('admin/structure/eav/attributes/' . $attribute->id() . '/formatter');
    $this->dontSeeErrorMessage();
    $web_assert->fieldValueEquals('fields[decimal_value][settings_edit_form][settings][thousand_separator]', $attribute_values['value_formatter_settings']['thousand_separator']);
    $web_assert->fieldValueEquals('fields[decimal_value][settings_edit_form][settings][decimal_separator]', $attribute_values['value_formatter_settings']['decimal_separator']);
    $web_assert->fieldValueEquals('fields[decimal_value][settings_edit_form][settings][scale]', $attribute_values['value_formatter_settings']['scale']);
    $web_assert->checkboxChecked('fields[decimal_value][settings_edit_form][settings][prefix_suffix]');
  }

  private function _testAddBooleanAttributeFromCode() {
    $web_assert = $this->assertSession();
    $this->drupalLoginAsRoot();

    $attribute_values = [
      'label' => 'Boolean attribute',
      'machine_name' => 'boolean_attr',
      'value_type' => 'boolean',
      'value_storage_config' => [
        'cardinality' => 1,
        'settings' => [],
      ],
      'value_config' => [
        'description' => '',
        'required' => FALSE,
        'default_value' => [
          0 => [
            'value' => 0,
          ],
        ],
        'settings' => [
          'on_label' => 'Yes',
          'off_label' => 'No',
        ],
      ],
      'value_widget_type' => 'boolean_checkbox',
      'value_widget_settings' => [
        'display_label' => TRUE,
      ],
      'value_formatter_type' => 'boolean',
      'value_formatter_settings' => [
        'format' => 'default',
        'format_custom_true' => '',
        'format_custom_false' => '',
      ],
      'description' => '',
      'category' => [1],
    ];
    $attribute = EavAttribute::create($attribute_values); /** @var EavAttribute $attribute */
    $attribute->save();

    // Attributes list
    $this->drupalGet('admin/structure/eav/attributes');
    $this->dontSeeErrorMessage();
    $web_assert->elementExists('css', 'tr[data-drupal-selector="edit-eav-attributes-' . $attribute->id() . '"]');

    // Edit attribute form
    $this->drupalGet('admin/structure/eav/attributes/' . $attribute->id() . '/edit');
    $this->dontSeeErrorMessage();
    $web_assert->fieldValueEquals('label[0][value]', $attribute_values['label']);
    $web_assert->fieldValueEquals('machine_name[0][value]', $attribute_values['machine_name']);
    $web_assert->fieldValueEquals('value_type', $attribute_values['value_type']);
    $web_assert->fieldValueEquals('value_widget_type', $attribute_values['value_widget_type']);
    $web_assert->fieldValueEquals('value_formatter_type', $attribute_values['value_formatter_type']);
    $web_assert->fieldValueEquals('category[0][target_id]', 'Category 1 (1)');
    $web_assert->fieldValueEquals('description[0][value]', $attribute_values['description']);

    // Storage settings form
    $this->drupalGet('admin/structure/eav/attributes/' . $attribute->id() . '/storage');
    $this->dontSeeErrorMessage();
    $web_assert->fieldValueEquals('cardinality', 'number');
    $web_assert->fieldValueEquals('cardinality_number', '1');

    // Field settings form
    $this->drupalGet('admin/structure/eav/attributes/' . $attribute->id() . '/field');
    $this->dontSeeErrorMessage();
    $web_assert->fieldValueEquals('settings[on_label]', $attribute_values['value_config']['settings']['on_label']);
    $web_assert->fieldValueEquals('settings[off_label]', $attribute_values['value_config']['settings']['off_label']);
    $web_assert->checkboxNotChecked('default_value_input[boolean_value][value]');

    // Widget settings form
    $this->drupalGet('admin/structure/eav/attributes/' . $attribute->id() . '/widget');
    $this->dontSeeErrorMessage();
    $web_assert->checkboxChecked('fields[boolean_value][settings_edit_form][settings][display_label]');

    // Formatter settings form
    $this->drupalGet('admin/structure/eav/attributes/' . $attribute->id() . '/formatter');
    $this->dontSeeErrorMessage();
    $web_assert->fieldValueEquals('fields[boolean_value][settings_edit_form][settings][format]', 'default');
  }

  private function _testAddEntityReferenceAttribute() {
    $web_assert = $this->assertSession();
    $this->drupalLoginAsRoot();

    $attribute_values = [
      'label' => 'Entity reference attribute',
      'machine_name' => 'entity_reference_attr',
      'value_type' => 'entity_reference',
      'value_widget_type' => 'options_select',
      'value_formatter_type' => 'entity_reference_label',
      'description' => '',
      'category' => [1],
    ];
    $attribute = EavAttribute::create($attribute_values); /** @var EavAttribute $attribute */
    $attribute->save();

    // Attributes list
    $this->drupalGet('admin/structure/eav/attributes');
    $this->dontSeeErrorMessage();
    $web_assert->elementExists('css', 'tr[data-drupal-selector="edit-eav-attributes-' . $attribute->id() . '"]');

    // Storage settings form
    $this->drupalGet('admin/structure/eav/attributes/' . $attribute->id() . '/storage');
    $this->dontSeeErrorMessage();
    $web_assert->fieldExists('settings[target_type]');
    $web_assert->fieldExists('cardinality');
    $web_assert->fieldExists('cardinality_number');
    $this->submitForm([
      'settings[target_type]' => 'taxonomy_term',
    ], 'Save field settings');
    $this->dontSeeErrorMessage();
    $web_assert->fieldValueEquals('settings[target_type]', 'taxonomy_term');
    $web_assert->fieldValueEquals('cardinality', 'number');
    $web_assert->fieldValueEquals('cardinality_number', '1');

    // Field settings form
    $this->drupalGet('admin/structure/eav/attributes/' . $attribute->id() . '/field');
    $this->dontSeeErrorMessage();
    $web_assert->fieldExists('settings[handler]');
    $web_assert->fieldExists('settings[handler_settings][target_bundles][category]');
    $web_assert->fieldExists('default_value_input[entity_reference_value][0][target_id]');
    $this->submitForm([
      'settings[handler_settings][target_bundles][category]' => 'category',
    ], 'Save settings');
    $this->dontSeeErrorMessage();
    $web_assert->fieldValueEquals('settings[handler]', 'default:taxonomy_term');
    $web_assert->checkboxNotChecked('settings[handler_settings][auto_create]');
    $web_assert->checkboxChecked('settings[handler_settings][target_bundles][category]');
    $web_assert->fieldValueEquals('default_value_input[entity_reference_value][0][target_id]', '');

    // Widget settings form
    $this->drupalGet('admin/structure/eav/attributes/' . $attribute->id() . '/widget');
    $this->dontSeeErrorMessage();
    $web_assert->pageTextContains('No settings.');

    // Formatter settings form
    $this->drupalGet('admin/structure/eav/attributes/' . $attribute->id() . '/formatter');
    $this->dontSeeErrorMessage();
    $web_assert->fieldExists('fields[entity_reference_value][settings_edit_form][settings][link]');
    $web_assert->checkboxChecked('fields[entity_reference_value][settings_edit_form][settings][link]');
    $this->submitForm([
      'fields[entity_reference_value][settings_edit_form][settings][link]' => FALSE,
    ], 'Save');
    $this->dontSeeErrorMessage();
    $web_assert->fieldExists('fields[entity_reference_value][settings_edit_form][settings][link]');
    $web_assert->checkboxNotChecked('fields[entity_reference_value][settings_edit_form][settings][link]');
  }

  private function _testAddListStringAttributeFromCode() {
    $web_assert = $this->assertSession();
    $this->drupalLoginAsRoot();

    $attribute_values = [
      'label' => 'List string attribute',
      'machine_name' => 'list_string_attr',
      'value_type' => 'list_string',
      'value_storage_config' => [
        'cardinality' => 1,
        'settings' => [
          'allowed_values' => [
            'foo' => 'Foo',
            'bar' => 'Bar',
          ],
          'allowed_values_function' => '',
        ],
      ],
      'value_config' => [
        'description' => '',
        'required' => FALSE,
        'default_value' => [
          0 => [
            'value' => 'bar',
          ],
        ],
        'settings' => [],
      ],
      'value_widget_type' => 'options_select',
      'value_widget_settings' => [],
      'value_formatter_type' => 'list_default',
      'value_formatter_settings' => [],
      'description' => '',
      'category' => [1],
    ];
    $attribute = EavAttribute::create($attribute_values); /** @var EavAttribute $attribute */
    $attribute->save();

    // Attributes list
    $this->drupalGet('admin/structure/eav/attributes');
    $this->dontSeeErrorMessage();
    $web_assert->elementExists('css', 'tr[data-drupal-selector="edit-eav-attributes-' . $attribute->id() . '"]');

    // Edit attribute form
    $this->drupalGet('admin/structure/eav/attributes/' . $attribute->id() . '/edit');
    $this->dontSeeErrorMessage();
    $web_assert->fieldValueEquals('label[0][value]', $attribute_values['label']);
    $web_assert->fieldValueEquals('machine_name[0][value]', $attribute_values['machine_name']);
    $web_assert->fieldValueEquals('value_type', $attribute_values['value_type']);
    $web_assert->fieldValueEquals('value_widget_type', $attribute_values['value_widget_type']);
    $web_assert->fieldValueEquals('value_formatter_type', $attribute_values['value_formatter_type']);
    $web_assert->fieldValueEquals('category[0][target_id]', 'Category 1 (1)');
    $web_assert->fieldValueEquals('description[0][value]', $attribute_values['description']);

    // Storage settings form
    $this->drupalGet('admin/structure/eav/attributes/' . $attribute->id() . '/storage');
    $this->dontSeeErrorMessage();
    $web_assert->fieldValueEquals('cardinality', 'number');
    $web_assert->fieldValueEquals('cardinality_number', '1');
    $web_assert->fieldValueEquals('settings[allowed_values]', "foo|Foo\nbar|Bar");

    // Field settings form
    $this->drupalGet('admin/structure/eav/attributes/' . $attribute->id() . '/field');
    $this->dontSeeErrorMessage();
    $web_assert->fieldExists('default_value_input[list_string_value]');
    $web_assert->fieldValueEquals('default_value_input[list_string_value]', $attribute_values['value_config']['default_value'][0]['value']);

    // Widget settings form
    $this->drupalGet('admin/structure/eav/attributes/' . $attribute->id() . '/widget');
    $this->dontSeeErrorMessage();
    $web_assert->pageTextContains('No settings.');

    // Formatter settings form
    $this->drupalGet('admin/structure/eav/attributes/' . $attribute->id() . '/formatter');
    $this->dontSeeErrorMessage();
    $web_assert->pageTextContains('No settings.');
  }

  private function _testAddListIntegerAttributeFromCode() {
    $web_assert = $this->assertSession();
    $this->drupalLoginAsRoot();

    $attribute_values = [
      'label' => 'List integer attribute',
      'machine_name' => 'list_integer_attr',
      'value_type' => 'list_integer',
      'value_storage_config' => [
        'cardinality' => 1,
        'settings' => [
          'allowed_values' => [
            1 => 'Foo',
            2 => 'Bar',
          ],
          'allowed_values_function' => '',
        ],
      ],
      'value_config' => [
        'description' => '',
        'required' => FALSE,
        'default_value' => [],
        'settings' => [],
      ],
      'value_widget_type' => 'options_select',
      'value_widget_settings' => [],
      'value_formatter_type' => 'list_default',
      'value_formatter_settings' => [],
      'description' => '',
      'category' => [1],
    ];
    $attribute = EavAttribute::create($attribute_values); /** @var EavAttribute $attribute */
    $attribute->save();

    // Attributes list
    $this->drupalGet('admin/structure/eav/attributes');
    $this->dontSeeErrorMessage();
    $web_assert->elementExists('css', 'tr[data-drupal-selector="edit-eav-attributes-' . $attribute->id() . '"]');

    // Edit attribute form
    $this->drupalGet('admin/structure/eav/attributes/' . $attribute->id() . '/edit');
    $this->dontSeeErrorMessage();
    $web_assert->fieldValueEquals('label[0][value]', $attribute_values['label']);
    $web_assert->fieldValueEquals('machine_name[0][value]', $attribute_values['machine_name']);
    $web_assert->fieldValueEquals('value_type', $attribute_values['value_type']);
    $web_assert->fieldValueEquals('value_widget_type', $attribute_values['value_widget_type']);
    $web_assert->fieldValueEquals('value_formatter_type', $attribute_values['value_formatter_type']);
    $web_assert->fieldValueEquals('category[0][target_id]', 'Category 1 (1)');
    $web_assert->fieldValueEquals('description[0][value]', $attribute_values['description']);

    // Storage settings form
    $this->drupalGet('admin/structure/eav/attributes/' . $attribute->id() . '/storage');
    $this->dontSeeErrorMessage();
    $web_assert->fieldValueEquals('cardinality', 'number');
    $web_assert->fieldValueEquals('cardinality_number', '1');
    $web_assert->fieldValueEquals('settings[allowed_values]', "1|Foo\n2|Bar");

    // Field settings form
    $this->drupalGet('admin/structure/eav/attributes/' . $attribute->id() . '/field');
    $this->dontSeeErrorMessage();
    $web_assert->fieldExists('default_value_input[list_integer_value]');

    // Widget settings form
    $this->drupalGet('admin/structure/eav/attributes/' . $attribute->id() . '/widget');
    $this->dontSeeErrorMessage();
    $web_assert->pageTextContains('No settings.');

    // Formatter settings form
    $this->drupalGet('admin/structure/eav/attributes/' . $attribute->id() . '/formatter');
    $this->dontSeeErrorMessage();
    $web_assert->pageTextContains('No settings.');
  }

  private function _testAttributesListFormSubmit() {
    $web_assert = $this->assertSession();
    $this->drupalLoginAsRoot();
    $list_string_attribute = EavAttribute::loadByMachineName('list_string_attr');
    $list_integer_attribute = EavAttribute::loadByMachineName('list_integer_attr');

    $this->drupalGet('admin/structure/eav/attributes');
    $this->dontSeeErrorMessage();
    $web_assert->fieldValueEquals('eav_attributes[' . $list_string_attribute->id() . '][weight]', $list_string_attribute->getWeight());
    $web_assert->fieldValueEquals('eav_attributes[' . $list_integer_attribute->id() . '][weight]', $list_integer_attribute->getWeight());

    $this->submitForm([
      'eav_attributes[' . $list_string_attribute->id() . '][weight]' => 100,
      'eav_attributes[' . $list_integer_attribute->id() . '][weight]' => 99,
    ], 'Save');
    $this->dontSeeErrorMessage();
    $web_assert->fieldValueEquals('eav_attributes[' . $list_string_attribute->id() . '][weight]', 100);
    $web_assert->fieldValueEquals('eav_attributes[' . $list_integer_attribute->id() . '][weight]', 99);
  }

  private function _testAttributesValuesFormBuilder() {
    $web_assert = $this->assertSession();
    $this->drupalLoginAsRoot();

    $this->drupalGet(Url::fromRoute('node.add', ['node_type' => 'product']));
    $this->dontSeeErrorMessage();
    $this->submitForm([
      'title[0][value]' => 'Test product',
      'field_category' => 1,
    ], 'Save');
    $this->dontSeeErrorMessage();
    $node_id = $this->getLastAddedNodeId('product');

    $this->drupalGet(Url::fromRoute('entity.node.canonical', ['node' => $node_id]));
    $this->dontSeeErrorMessage();
    $edit_eav_attributes_page_url = Url::fromRoute('entity.node.edit_eav_attributes', ['node' => $node_id]);
    $web_assert->elementExists('css', 'ul.primary a[href="' . $edit_eav_attributes_page_url->toString() . '"]');

    $this->drupalGet($edit_eav_attributes_page_url);
    $this->dontSeeErrorMessage();

    // String
    $string_attribute = EavAttribute::loadByMachineName('string_attr');
    $string_attribute_value_storage_config_array = $string_attribute->getValueStorageConfigArray();
    $string_attribute_value_config_array = $string_attribute->getValueConfigArray();
    $string_attribute_value_widget_settings = $string_attribute->getValueWidgetSettings();
    $web_assert->elementExists('css', 'input[type="text"][name="field_eav[string_attr][string_value][0][value]"]');
    $web_assert->fieldValueEquals('field_eav[string_attr][string_value][0][value]', $string_attribute_value_config_array['default_value'][0]['value']);
    $web_assert->elementTextContains('css', '#edit-field-eav-string-attr-string-value-0-value--description', $string_attribute_value_config_array['description']);
    $web_assert->elementAttributeContains('css', '#edit-field-eav-string-attr-string-value-0-value', 'size', $string_attribute_value_widget_settings['size']);
    $web_assert->elementAttributeContains('css', '#edit-field-eav-string-attr-string-value-0-value', 'maxlength', $string_attribute_value_storage_config_array['settings']['max_length']);
    $web_assert->elementAttributeContains('css', '#edit-field-eav-string-attr-string-value-0-value', 'placeholder', $string_attribute_value_widget_settings['placeholder']);

    // Second string
    $web_assert->elementExists('css', 'input[type="text"][name="field_eav[second_string_attr][string_value][0][value]"]');
    $web_assert->fieldValueEquals('field_eav[second_string_attr][string_value][0][value]', '');
    $web_assert->elementExists('css', '.form-item-field-eav-second-string-attr-string-value-0-value');
    $web_assert->elementNotExists('css', '.form-item-field-eav-second-string-attr-string-value-0-value .description');

    // String long
    $string_long_attribute = EavAttribute::loadByMachineName('string_long_attr');
    $string_long_attribute_widget_settings = $string_long_attribute->getValueWidgetSettings();
    $web_assert->elementExists('css', 'textarea[name="field_eav[string_long_attr][string_long_value][0][value]"]');
    $web_assert->fieldValueEquals('field_eav[string_long_attr][string_long_value][0][value]', '');
    $web_assert->elementAttributeContains('css', '#edit-field-eav-string-long-attr-string-long-value-0-value', 'rows', $string_long_attribute_widget_settings['rows']);
    $web_assert->elementAttributeContains('css', '#edit-field-eav-string-long-attr-string-long-value-0-value', 'placeholder', $string_long_attribute_widget_settings['placeholder']);

    // Integer
    $integer_attribute = EavAttribute::loadByMachineName('integer_attr');
    $integer_attribute_value_config = $integer_attribute->getValueConfigArray();
    $web_assert->elementExists('css', 'input[type="number"][name="field_eav[integer_attr][integer_value][0][value]"]');
    $web_assert->elementTextContains('css', '.form-item-field-eav-integer-attr-integer-value-0-value .field-suffix', $integer_attribute_value_config['settings']['suffix']);
    $web_assert->elementNotExists('css', '.form-item-field-eav-integer-attr-integer-value-0-value .field-prefix');

    // Decimal
    $decimal_attribute = EavAttribute::loadByMachineName('decimal_attr');
    $decimal_attribute_value_config = $decimal_attribute->getValueConfigArray();
    $web_assert->elementExists('css', 'input[type="number"][name="field_eav[decimal_attr][decimal_value][0][value]"]');
    $web_assert->elementAttributeContains('css', '#edit-field-eav-decimal-attr-decimal-value-0-value', 'min', $decimal_attribute_value_config['settings']['min']);
    $web_assert->elementAttributeContains('css', '#edit-field-eav-decimal-attr-decimal-value-0-value', 'max', $decimal_attribute_value_config['settings']['max']);
    $web_assert->elementTextContains('css', '.form-item-field-eav-decimal-attr-decimal-value-0-value .field-prefix', $decimal_attribute_value_config['settings']['prefix']);
    $web_assert->elementNotExists('css', '.form-item-field-eav-decimal-attr-decimal-value-0-value .field-suffix');

    // Boolean
    $web_assert->elementExists('css', 'input[type="checkbox"][name="field_eav[boolean_attr][boolean_value][value]"]');
    $web_assert->checkboxNotChecked('field_eav[boolean_attr][boolean_value][value]');

    // Entity reference
    $web_assert->elementExists('css', 'select[name="field_eav[entity_reference_attr][entity_reference_value]"]');
    $web_assert->elementExists('css', '#edit-field-eav-entity-reference-attr-entity-reference-value option[value="1"]');

    // List integer
    $web_assert->elementExists('css', 'select[name="field_eav[list_integer_attr][list_integer_value]"]');
    $web_assert->elementExists('css', '#edit-field-eav-list-integer-attr-list-integer-value option[value="1"]');
    $web_assert->elementExists('css', '#edit-field-eav-list-integer-attr-list-integer-value option[value="2"]');

    // List string
    $web_assert->elementExists('css', 'select[name="field_eav[list_string_attr][list_string_value]"]');
    $web_assert->elementExists('css', '#edit-field-eav-list-string-attr-list-string-value option[value="foo"]');
    $web_assert->elementExists('css', '#edit-field-eav-list-string-attr-list-string-value option[value="bar"]');
    $web_assert->fieldValueEquals('field_eav[list_string_attr][list_string_value]', 'bar');

    $web_assert->elementExists('css', '.eav-entity-attributes-form .form-actions .form-submit');
  }

  private function _testAttributesValuesFormSubmit() {
    $web_assert = $this->assertSession();
    $this->drupalLoginAsRoot();

    $node_id = $this->getLastAddedNodeId('product');
    $this->drupalGet('node/' . $node_id . '/edit/eav-attributes');
    $this->dontSeeErrorMessage();

    $form_values = [
      'field_eav[string_attr][string_value][0][value]' => 'String attribute value',
      'field_eav[string_long_attr][string_long_value][0][value]' => 'Long string attribute value',
      'field_eav[integer_attr][integer_value][0][value]' => 10,
      'field_eav[decimal_attr][decimal_value][0][value]' => 15.9,
      'field_eav[boolean_attr][boolean_value][value]' => TRUE,
      'field_eav[entity_reference_attr][entity_reference_value]' => 4,
      'field_eav[list_integer_attr][list_integer_value]' => 1,
      'field_eav[list_string_attr][list_string_value]' => 'foo',
    ];
    $this->submitForm($form_values, 'Save');
    $this->dontSeeErrorMessage();
    $this->seeSystemMessage('Values saved');

    $web_assert->fieldValueEquals('field_eav[string_attr][string_value][0][value]', $form_values['field_eav[string_attr][string_value][0][value]']);
    $web_assert->fieldValueEquals('field_eav[string_long_attr][string_long_value][0][value]', $form_values['field_eav[string_long_attr][string_long_value][0][value]']);
    $web_assert->fieldValueEquals('field_eav[integer_attr][integer_value][0][value]', $form_values['field_eav[integer_attr][integer_value][0][value]']);
    $web_assert->fieldValueEquals('field_eav[decimal_attr][decimal_value][0][value]', '15.90');
    $web_assert->checkboxChecked('field_eav[boolean_attr][boolean_value][value]');
    $web_assert->fieldValueEquals('field_eav[entity_reference_attr][entity_reference_value]', $form_values['field_eav[entity_reference_attr][entity_reference_value]']);
    $web_assert->fieldValueEquals('field_eav[list_integer_attr][list_integer_value]', $form_values['field_eav[list_integer_attr][list_integer_value]']);
    $web_assert->fieldValueEquals('field_eav[list_string_attr][list_string_value]', $form_values['field_eav[list_string_attr][list_string_value]']);
  }

  private function _testEavFieldFormatter() {
    $web_assert = $this->assertSession();
    $this->drupalLoginAsRoot();

    $node_id = $this->getLastAddedNodeId('product');
    $this->drupalGet(Url::fromRoute('entity.node.canonical', ['node' => $node_id]));
    $this->dontSeeErrorMessage();

    $web_assert->elementTextContains('css', '.eav-list__item--string-attr .eav-list__attribute-label', 'String attribute');
    $web_assert->elementTextContains('css', '.eav-list__item--string-attr .eav-list__value', 'String attribute value');

    $web_assert->elementNotExists('css', '.eav-list__item--second-string-attr');

    $web_assert->elementTextContains('css', '.eav-list__item--string-long-attr .eav-list__attribute-label', 'Long string attribute');
    $web_assert->elementTextContains('css', '.eav-list__item--string-long-attr .eav-list__value', 'Long string attribute value');

    $web_assert->elementTextContains('css', '.eav-list__item--integer-attr .eav-list__value', '10sffx');

    $web_assert->elementTextContains('css', '.eav-list__item--decimal-attr .eav-list__value', '$15.90');

    $web_assert->elementTextContains('css', '.eav-list__item--boolean-attr .eav-list__value', 'Yes');

    $web_assert->elementTextContains('css', '.eav-list__item--entity-reference-attr .eav-list__value', 'Category 1.1');

    $web_assert->elementTextContains('css', '.eav-list__item--list-integer-attr .eav-list__value', 'Foo');

    $web_assert->elementTextContains('css', '.eav-list__item--list-string-attr .eav-list__value', 'Foo');
  }

  private function _testEavFieldFormatterWithLayoutBuilder() {
    $web_assert = $this->assertSession();
    $this->drupalLoginAsRoot();

    // Install layout modules
    $module_installer = $this->container->get('module_installer');
    $module_installer->install(['layout_builder', 'layout_discovery']);

    // Turn on layout builder to product node bundle
    $this->drupalGet('admin/structure/types/manage/product/display');
    $this->dontSeeErrorMessage();
    $this->submitForm([
      'layout[enabled]' => TRUE,
    ], 'Save');
    $this->dontSeeErrorMessage();
    $web_assert->elementExists('css', 'a#edit-manage-layout');

    // Test formatter
    $node_id = $this->getLastAddedNodeId('product');
    $this->drupalGet(Url::fromRoute('entity.node.canonical', ['node' => $node_id]));
    $this->dontSeeErrorMessage();
    $web_assert->elementExists('css', 'table.eav-list');

    // Clean environment
    $module_installer->uninstall(['layout_builder', 'layout_discovery']);
  }

  private function _testAttributeStorageSettingsFormAfterAddedEavValueEntities() {
    $web_assert = $this->assertSession();
    $this->drupalLoginAsRoot();

    $string_attribute = EavAttribute::loadByMachineName('string_attr');
    $this->drupalGet(Url::fromRoute('entity.eav_attribute.value_storage_form', ['eav_attribute' => $string_attribute->id()]));
    $this->dontSeeErrorMessage();
    $web_assert->elementExists('css', '#edit-settings-max-length');
    $web_assert->elementNotExists('css', '#edit-settings-max-length[disabled]');
  }

  private function _testAttributesValuesFormSetNewValues() {
    $web_assert = $this->assertSession();
    $this->drupalLoginAsRoot();

    $node_id = $this->getLastAddedNodeId('product');
    $node = Node::load($node_id);
    $eav_value_ids = $this->getEavFieldItems($node)->getValueEntitiesIds();
    $this->assertGreaterThan(0, count($eav_value_ids));

    $this->drupalGet(Url::fromRoute('entity.node.edit_eav_attributes', ['node' => $node_id]));
    $this->dontSeeErrorMessage();

    $form_values = [
      'field_eav[string_attr][string_value][0][value]' => 'New string attribute value',
      'field_eav[second_string_attr][string_value][0][value]' => 'New second string attribute value',
      'field_eav[string_long_attr][string_long_value][0][value]' => 'New long string attribute value',
      'field_eav[integer_attr][integer_value][0][value]' => '11',
      'field_eav[decimal_attr][decimal_value][0][value]' => '22',
      'field_eav[boolean_attr][boolean_value][value]' => TRUE,
      'field_eav[entity_reference_attr][entity_reference_value]' => '1',
      'field_eav[list_integer_attr][list_integer_value]' => '2',
      'field_eav[list_string_attr][list_string_value]' => 'bar',
    ];
    $this->submitForm($form_values, 'Save');
    $this->dontSeeErrorMessage();
    $this->seeSystemMessage('Values saved');

    $web_assert->fieldValueEquals('field_eav[string_attr][string_value][0][value]', $form_values['field_eav[string_attr][string_value][0][value]']);
    $web_assert->fieldValueEquals('field_eav[string_long_attr][string_long_value][0][value]', $form_values['field_eav[string_long_attr][string_long_value][0][value]']);
    $web_assert->fieldValueEquals('field_eav[integer_attr][integer_value][0][value]', $form_values['field_eav[integer_attr][integer_value][0][value]']);
    $web_assert->fieldValueEquals('field_eav[decimal_attr][decimal_value][0][value]', '22.00');
    $web_assert->checkboxChecked('field_eav[boolean_attr][boolean_value][value]');
    $web_assert->fieldValueEquals('field_eav[entity_reference_attr][entity_reference_value]', $form_values['field_eav[entity_reference_attr][entity_reference_value]']);
    $web_assert->fieldValueEquals('field_eav[list_integer_attr][list_integer_value]', $form_values['field_eav[list_integer_attr][list_integer_value]']);
    $web_assert->fieldValueEquals('field_eav[list_string_attr][list_string_value]', $form_values['field_eav[list_string_attr][list_string_value]']);

    $this->resetEntityStorageCaches();
    $node = Node::load($node_id);
    $this->assertIdentical([], array_diff($eav_value_ids, $this->getEavFieldItems($node)->getValueEntitiesIds()));
  }

  private function _testAttributesValuesFormSetEmptyValues() {
    $web_assert = $this->assertSession();
    $this->drupalLoginAsRoot();

    $node_id = $this->getLastAddedNodeId('product');
    $this->drupalGet(Url::fromRoute('entity.node.edit_eav_attributes', ['node' => $node_id]));
    $this->dontSeeErrorMessage();

    $form_values = [
      'field_eav[string_attr][string_value][0][value]' => '',
      'field_eav[string_long_attr][string_long_value][0][value]' => '',
      'field_eav[integer_attr][integer_value][0][value]' => '',
      'field_eav[decimal_attr][decimal_value][0][value]' => '',
      'field_eav[boolean_attr][boolean_value][value]' => FALSE,
      'field_eav[entity_reference_attr][entity_reference_value]' => '_none',
      'field_eav[list_integer_attr][list_integer_value]' => '_none',
      'field_eav[list_string_attr][list_string_value]' => '_none',
    ];
    $this->submitForm($form_values, 'Save');
    $this->dontSeeErrorMessage();
    $this->seeSystemMessage('Values saved');

    $web_assert->fieldValueEquals('field_eav[string_attr][string_value][0][value]', '');
    $web_assert->fieldValueEquals('field_eav[string_long_attr][string_long_value][0][value]', '');
    $web_assert->fieldValueEquals('field_eav[integer_attr][integer_value][0][value]', '');
    $web_assert->fieldValueEquals('field_eav[decimal_attr][decimal_value][0][value]', '');
    $web_assert->checkboxNotChecked('field_eav[boolean_attr][boolean_value][value]');
    $web_assert->fieldValueEquals('field_eav[entity_reference_attr][entity_reference_value]', '_none');
    $web_assert->fieldValueEquals('field_eav[list_integer_attr][list_integer_value]', '_none');
    $web_assert->fieldValueEquals('field_eav[list_string_attr][list_string_value]', '_none');
  }

  private function _testAttributesValuesFormClearButton() {
    $web_assert = $this->assertSession();
    $this->drupalLoginAsRoot();

    $node_id = $this->getLastAddedNodeId('product');
    $node = Node::load($node_id);
    $eav_value_ids = $this->getEavFieldItems($node)->getValueEntitiesIds();
    $this->assertGreaterThan(0, count($eav_value_ids));

    $this->drupalGet(Url::fromRoute('entity.node.edit_eav_attributes', ['node' => $node_id]));
    $this->dontSeeErrorMessage();
    $web_assert->elementExists('css', 'input[type="submit"][value="Clear all attributes value"]');

    $this->click('#edit-delete');
    $this->dontSeeErrorMessage();
    $this->seeSystemMessage('Values cleared');
    $web_assert->fieldValueEquals('field_eav[second_string_attr][string_value][0][value]', '');
    $web_assert->fieldValueEquals('field_eav[string_long_attr][string_long_value][0][value]', '');
    $web_assert->checkboxNotChecked('field_eav[boolean_attr][boolean_value][value]');

    $this->resetEntityStorageCaches();
    $node = Node::load($node_id);
    $this->assertIdentical(0, $this->getEavFieldItems($node)->count());
    foreach ($eav_value_ids as $eav_value_id) {
      $eav_value = EavValue::load($eav_value_id);
      $this->assertNull($eav_value);
    }
  }

  private function _testDeleteEavValueEntitiesAfterDeleteHostEntity() {
    // @TODO
  }

  private function _testDeleteEavValueEntitiesAfterDeleteAttribute() {
    // @TODO
  }

  private function _testDeleteEavValueEntitiesAfterDeleteEavField() {
    // @TODO
  }

  private function _testAttributesEditFormByCategory() {
    // @TODO
  }

  private function _testDeleteEavValueEntitiesAfterMoveAttributeFromCategory() {
    // @TODO
  }

  private function _testAttributesValuesFormAfterAddNewAttribute() {
    // @TODO
    // Set attributes values
    // Add new attribute
    // Test attributes form + form submit
  }

  private function _testAttributesValuesFormOnCommerceProduct() {
    $web_assert = $this->assertSession();
    $this->drupalLoginAsRoot();

    // Install commerce_product module
    $module_installer = $this->container->get('module_installer');
    $module_installer->install(['commerce_product']);

    // Create store
    $store = Store::create([
      'type' => 'online',
      'name' => 'Default store',
    ]);
    $store->save();

    // Add field to default product type
    $this->createField(
      'commerce_product',
      'default',
      'field_eav',
      'eav',
      'EAV attributes',
      [],
      [],
      'eav_widget',
      [],
      'eav_formatter',
      []
    );
    drupal_flush_all_caches();

    // Create product
    $product = Product::create([
      'type' => 'default',
      'title' => 'Product 1',
    ]);
    $product->save();

    $this->drupalGet(Url::fromRoute('entity.commerce_product.canonical', ['commerce_product' => $product->id()]));
    $this->dontSeeErrorMessage();
    $web_assert->elementTextContains('css', 'ul.primary', 'Edit EAV attributes');

    $this->drupalGet(Url::fromRoute('entity.commerce_product.edit_eav_attributes', ['commerce_product' => $product->id()]));
    $this->dontSeeErrorMessage();
    $web_assert->elementExists('css', 'form.eav-entity-attributes-form');
  }

}
